@@ -1,7 +1,7 @@
 /******************************************************************************
  * The MIT License (MIT)
  *
- * Copyright (c) 2019-2020 Baldur Karlsson
+ * Copyright (c) 2019-2021 Baldur Karlsson
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -22,6 +22,8 @@
  * THE SOFTWARE.
  ******************************************************************************/
 
+#define AMD_GPUPERFAPI_SKIP_VULKAN_INCLUDE 1
+
 #include <algorithm>
 #include <iterator>
 #include "vk_core.h"
@@ -29,7 +31,7 @@
 #include "vk_resources.h"
 
 #include "driver/ihv/amd/amd_counters.h"
-#include "driver/ihv/amd/official/GPUPerfAPI/Include/GPUPerfAPI-VK.h"
+#include "driver/ihv/amd/official/GPUPerfAPI/Include/gpu_perf_api_vk.h"
 #include "strings/string_utils.h"
 
 static uint32_t FromKHRCounter(GPUCounter counterID)
@@ -42,66 +44,32 @@ static GPUCounter ToKHRCounter(uint32_t idx)
   return (GPUCounter)((uint32_t)GPUCounter::FirstVulkanExtended + idx);
 }
 
-static void GetKHRUnitDescription(const VkPerformanceCounterUnitKHR khrUnit, CounterUnit &unit,
-                                  CompType &type, uint32_t &byteWidth)
+static bool isFloatKhrStorage(const VkPerformanceCounterStorageKHR khrStorage)
 {
+  return khrStorage == VK_PERFORMANCE_COUNTER_STORAGE_FLOAT32_KHR ||
+         khrStorage == VK_PERFORMANCE_COUNTER_STORAGE_FLOAT64_KHR;
+}
+
+static void GetKHRUnitDescription(const VkPerformanceCounterUnitKHR khrUnit,
+                                  const VkPerformanceCounterStorageKHR khrStorage,
+                                  CounterUnit &unit, CompType &type, uint32_t &byteWidth)
+{
+  type = isFloatKhrStorage(khrStorage) ? CompType::Float : CompType::UInt;
+  byteWidth = 8;
+
   switch(khrUnit)
   {
-    case VK_PERFORMANCE_COUNTER_UNIT_GENERIC_KHR:
-      unit = CounterUnit::Absolute;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_PERCENTAGE_KHR:
-      unit = CounterUnit::Percentage;
-      type = CompType::Double;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_NANOSECONDS_KHR:
-      unit = CounterUnit::Seconds;
-      type = CompType::Double;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_BYTES_KHR:
-      unit = CounterUnit::Bytes;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_BYTES_PER_SECOND_KHR:
-      unit = CounterUnit::Ratio;
-      type = CompType::Double;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_KELVIN_KHR:
-      unit = CounterUnit::Absolute;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_WATTS_KHR:
-      unit = CounterUnit::Absolute;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_VOLTS_KHR:
-      unit = CounterUnit::Absolute;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_AMPS_KHR:
-      unit = CounterUnit::Absolute;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_HERTZ_KHR:
-      unit = CounterUnit::Absolute;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
-    case VK_PERFORMANCE_COUNTER_UNIT_CYCLES_KHR:
-      unit = CounterUnit::Cycles;
-      type = CompType::UInt;
-      byteWidth = 8;
-      return;
+    case VK_PERFORMANCE_COUNTER_UNIT_GENERIC_KHR: unit = CounterUnit::Absolute; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_PERCENTAGE_KHR: unit = CounterUnit::Percentage; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_NANOSECONDS_KHR: unit = CounterUnit::Seconds; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_BYTES_KHR: unit = CounterUnit::Bytes; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_BYTES_PER_SECOND_KHR: unit = CounterUnit::Ratio; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_KELVIN_KHR: unit = CounterUnit::Absolute; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_WATTS_KHR: unit = CounterUnit::Absolute; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_VOLTS_KHR: unit = CounterUnit::Absolute; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_AMPS_KHR: unit = CounterUnit::Absolute; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_HERTZ_KHR: unit = CounterUnit::Absolute; return;
+    case VK_PERFORMANCE_COUNTER_UNIT_CYCLES_KHR: unit = CounterUnit::Cycles; return;
     default: RDCERR("Invalid performance counter unit %d", khrUnit);
   }
 }
@@ -114,7 +82,7 @@ void VulkanReplay::convertKhrCounterResult(CounterResult &rdcResult,
   CounterUnit unit;
   CompType type;
   uint32_t byteWidth;
-  GetKHRUnitDescription(khrUnit, unit, type, byteWidth);
+  GetKHRUnitDescription(khrUnit, khrStorage, unit, type, byteWidth);
 
   double value;
 
@@ -133,8 +101,15 @@ void VulkanReplay::convertKhrCounterResult(CounterResult &rdcResult,
   // Special case for time units, renderdoc only has a Seconds type.
   if(khrUnit == VK_PERFORMANCE_COUNTER_UNIT_NANOSECONDS_KHR)
   {
-    RDCASSERT(type == CompType::Double);
-    rdcResult.value.d /= 1000.0 * 1000.0 * 1000.0;
+    if((khrStorage == VK_PERFORMANCE_COUNTER_STORAGE_FLOAT64_KHR) ||
+       (khrStorage == VK_PERFORMANCE_COUNTER_STORAGE_FLOAT32_KHR))
+    {
+      rdcResult.value.d = rdcResult.value.d / (1000.0 * 1000.0 * 1000.0);
+    }
+    else
+    {
+      rdcResult.value.d = (double)(rdcResult.value.u64) / (1000.0 * 1000.0 * 1000.0);
+    }
   }
 }
 
@@ -142,7 +117,7 @@ rdcarray<GPUCounter> VulkanReplay::EnumerateCounters()
 {
   rdcarray<GPUCounter> ret;
 
-  VkPhysicalDeviceFeatures availableFeatures = m_pDriver->GetDeviceFeatures();
+  VkPhysicalDeviceFeatures availableFeatures = m_pDriver->GetDeviceEnabledFeatures();
 
   ret.push_back(GPUCounter::EventGPUDuration);
   if(availableFeatures.pipelineStatisticsQuery)
@@ -181,7 +156,12 @@ rdcarray<GPUCounter> VulkanReplay::EnumerateCounters()
         Unwrap(physDev), 0, &khrCounters, &m_KHRCounters[0], &m_KHRCountersDescriptions[0]);
 
     for(uint32_t c = 0; c < khrCounters; c++)
-      ret.push_back(ToKHRCounter(c));
+    {
+      // Only report counters with command scope. We currently don't
+      // have use for renderpass ones.
+      if(m_KHRCounters[c].scope == VK_PERFORMANCE_COUNTER_SCOPE_COMMAND_KHR)
+        ret.push_back(ToKHRCounter(c));
+    }
   }
 
   if(m_pAMDCounters)
@@ -226,7 +206,12 @@ CounterDescription VulkanReplay::DescribeCounter(GPUCounter counterID)
     desc.uuid.words[2] = uuid_dwords[2];
     desc.uuid.words[3] = uuid_dwords[3];
 
-    GetKHRUnitDescription(khrCounter.unit, rdcDesc.unit, rdcDesc.resultType, rdcDesc.resultByteWidth);
+    GetKHRUnitDescription(khrCounter.unit, khrCounter.storage, rdcDesc.unit, rdcDesc.resultType,
+                          rdcDesc.resultByteWidth);
+
+    // Special chase for time units.
+    if(khrCounter.unit == VK_PERFORMANCE_COUNTER_UNIT_NANOSECONDS_KHR)
+      rdcDesc.resultType = CompType::Float;
 
     return rdcDesc;
   }
@@ -246,7 +231,7 @@ CounterDescription VulkanReplay::DescribeCounter(GPUCounter counterID)
       desc.description =
           "Time taken for this event on the GPU, as measured by delta between two GPU timestamps.";
       desc.resultByteWidth = 8;
-      desc.resultType = CompType::Double;
+      desc.resultType = CompType::Float;
       desc.unit = CounterUnit::Seconds;
       break;
     case GPUCounter::InputVerticesRead:
@@ -345,16 +330,16 @@ CounterDescription VulkanReplay::DescribeCounter(GPUCounter counterID)
   return desc;
 }
 
-struct VulkanAMDDrawCallback : public VulkanDrawcallCallback
+struct VulkanAMDActionCallback : public VulkanActionCallback
 {
-  VulkanAMDDrawCallback(WrappedVulkan *dev, VulkanReplay *rp, uint32_t &sampleIndex,
-                        rdcarray<uint32_t> &eventIDs)
+  VulkanAMDActionCallback(WrappedVulkan *dev, VulkanReplay *rp, uint32_t &sampleIndex,
+                          rdcarray<uint32_t> &eventIDs)
       : m_pDriver(dev), m_pReplay(rp), m_pSampleId(&sampleIndex), m_pEventIds(&eventIDs)
   {
-    m_pDriver->SetDrawcallCB(this);
+    m_pDriver->SetActionCB(this);
   }
 
-  virtual ~VulkanAMDDrawCallback() { m_pDriver->SetDrawcallCB(NULL); }
+  virtual ~VulkanAMDActionCallback() { m_pDriver->SetActionCB(NULL); }
   void PreDraw(uint32_t eid, VkCommandBuffer cmd) override
   {
     m_pEventIds->push_back(eid);
@@ -399,13 +384,22 @@ struct VulkanAMDDrawCallback : public VulkanDrawcallCallback
   void PreDispatch(uint32_t eid, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
   bool PostDispatch(uint32_t eid, VkCommandBuffer cmd) override { return PostDraw(eid, cmd); }
   void PostRedispatch(uint32_t eid, VkCommandBuffer cmd) override { PostRedraw(eid, cmd); }
-  void PreMisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
-  bool PostMisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override
+  void PreMisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
   {
+    if(flags & ActionFlags::PassBoundary)
+      return;
+    PreDraw(eid, cmd);
+  }
+  bool PostMisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
+  {
+    if(flags & ActionFlags::PassBoundary)
+      return false;
     return PostDraw(eid, cmd);
   }
-  void PostRemisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override
+  void PostRemisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
   {
+    if(flags & ActionFlags::PassBoundary)
+      return;
     PostRedraw(eid, cmd);
   }
 
@@ -413,6 +407,15 @@ struct VulkanAMDDrawCallback : public VulkanDrawcallCallback
   {
     m_AliasEvents.push_back(make_rdcpair(primary, alias));
   }
+  bool SplitSecondary() override { return false; }
+  void PreCmdExecute(uint32_t baseEid, uint32_t secondaryFirst, uint32_t secondaryLast,
+                     VkCommandBuffer cmd) override
+  {
+  }
+  void PostCmdExecute(uint32_t baseEid, uint32_t secondaryFirst, uint32_t secondaryLast,
+                      VkCommandBuffer cmd) override
+  {
+  }
 
   uint32_t *m_pSampleId;
   WrappedVulkan *m_pDriver;
@@ -431,7 +434,7 @@ void VulkanReplay::FillTimersAMD(uint32_t *eventStartID, uint32_t *sampleIndex,
 {
   uint32_t maxEID = m_pDriver->GetMaxEID();
 
-  m_pAMDDrawCallback = new VulkanAMDDrawCallback(m_pDriver, this, *sampleIndex, *eventIDs);
+  m_pAMDActionCallback = new VulkanAMDActionCallback(m_pDriver, this, *sampleIndex, *eventIDs);
 
   // replay the events to perform all the queries
   m_pDriver->ReplayLog(*eventStartID, maxEID, eReplay_Full);
@@ -439,8 +442,8 @@ void VulkanReplay::FillTimersAMD(uint32_t *eventStartID, uint32_t *sampleIndex,
 
 rdcarray<CounterResult> VulkanReplay::FetchCountersAMD(const rdcarray<GPUCounter> &counters)
 {
-  GPA_vkContextOpenInfo context = {Unwrap(m_pDriver->GetInstance()),
-                                   Unwrap(m_pDriver->GetPhysDev()), Unwrap(m_pDriver->GetDev())};
+  GpaVkContextOpenInfo context = {Unwrap(m_pDriver->GetInstance()), Unwrap(m_pDriver->GetPhysDev()),
+                                  Unwrap(m_pDriver->GetDev())};
 
   if(!m_pAMDCounters->BeginMeasurementMode(AMDCounters::ApiType::Vk, (void *)&context))
   {
@@ -487,13 +490,13 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersAMD(const rdcarray<GPUCounter
   rdcarray<CounterResult> ret =
       m_pAMDCounters->GetCounterData(sessionID, sampleIndex, eventIDs, counters);
 
-  for(size_t i = 0; i < m_pAMDDrawCallback->m_AliasEvents.size(); i++)
+  for(size_t i = 0; i < m_pAMDActionCallback->m_AliasEvents.size(); i++)
   {
     for(size_t c = 0; c < counters.size(); c++)
     {
       CounterResult search;
       search.counter = counters[c];
-      search.eventId = m_pAMDDrawCallback->m_AliasEvents[i].first;
+      search.eventId = m_pAMDActionCallback->m_AliasEvents[i].first;
 
       // find the result we're aliasing
       int32_t idx = ret.indexOf(search);
@@ -501,7 +504,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersAMD(const rdcarray<GPUCounter
       {
         // duplicate the result and append
         CounterResult aliased = ret[idx];
-        aliased.eventId = m_pAMDDrawCallback->m_AliasEvents[i].second;
+        aliased.eventId = m_pAMDActionCallback->m_AliasEvents[i].second;
         ret.push_back(aliased);
       }
       else
@@ -512,7 +515,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersAMD(const rdcarray<GPUCounter
     }
   }
 
-  SAFE_DELETE(m_pAMDDrawCallback);
+  SAFE_DELETE(m_pAMDActionCallback);
 
   // sort so that the alias results appear in the right places
   std::sort(ret.begin(), ret.end());
@@ -522,14 +525,14 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersAMD(const rdcarray<GPUCounter
   return ret;
 }
 
-struct VulkanKHRCallback : public VulkanDrawcallCallback
+struct VulkanKHRCallback : public VulkanActionCallback
 {
   VulkanKHRCallback(WrappedVulkan *vk, VulkanReplay *rp, VkQueryPool qp)
       : m_pDriver(vk), m_pReplay(rp), m_QueryPool(qp)
   {
-    m_pDriver->SetDrawcallCB(this);
+    m_pDriver->SetActionCB(this);
   }
-  ~VulkanKHRCallback() { m_pDriver->SetDrawcallCB(NULL); }
+  ~VulkanKHRCallback() { m_pDriver->SetActionCB(NULL); }
   void PreDraw(uint32_t eid, VkCommandBuffer cmd) override
   {
     ObjDisp(cmd)->CmdBeginQuery(Unwrap(cmd), m_QueryPool, (uint32_t)m_Results.size(), 0);
@@ -547,12 +550,12 @@ struct VulkanKHRCallback : public VulkanDrawcallCallback
   void PreDispatch(uint32_t eid, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
   bool PostDispatch(uint32_t eid, VkCommandBuffer cmd) override { return PostDraw(eid, cmd); }
   void PostRedispatch(uint32_t eid, VkCommandBuffer cmd) override { PostRedraw(eid, cmd); }
-  void PreMisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
-  bool PostMisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override
+  void PreMisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
+  bool PostMisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
   {
     return PostDraw(eid, cmd);
   }
-  void PostRemisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override
+  void PostRemisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
   {
     PostRedraw(eid, cmd);
   }
@@ -560,6 +563,15 @@ struct VulkanKHRCallback : public VulkanDrawcallCallback
   {
     m_AliasEvents.push_back(std::make_pair(primary, alias));
   }
+  bool SplitSecondary() override { return false; }
+  void PreCmdExecute(uint32_t baseEid, uint32_t secondaryFirst, uint32_t secondaryLast,
+                     VkCommandBuffer cmd) override
+  {
+  }
+  void PostCmdExecute(uint32_t baseEid, uint32_t secondaryFirst, uint32_t secondaryLast,
+                      VkCommandBuffer cmd) override
+  {
+  }
 
   void PreEndCommandBuffer(VkCommandBuffer cmd) override {}
   WrappedVulkan *m_pDriver;
@@ -604,21 +616,24 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersKHR(const rdcarray<GPUCounter
 
   VkQueryPool queryPool;
   vkr = ObjDisp(dev)->CreateQueryPool(Unwrap(dev), &queryPoolCreateInfo, NULL, &queryPool);
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   // Reset query pool
   VkCommandBuffer cmd = m_pDriver->GetNextCmd();
 
+  if(cmd == VK_NULL_HANDLE)
+    return {};
+
   VkCommandBufferBeginInfo beginInfo = {VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO, NULL,
                                         VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT};
 
   vkr = ObjDisp(dev)->BeginCommandBuffer(Unwrap(cmd), &beginInfo);
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   ObjDisp(dev)->CmdResetQueryPool(Unwrap(cmd), queryPool, 0, maxEID);
 
   vkr = ObjDisp(dev)->EndCommandBuffer(Unwrap(cmd));
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   m_pDriver->SubmitCmds();
 
@@ -644,7 +659,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersKHR(const rdcarray<GPUCounter
       Unwrap(dev), queryPool, 0, (uint32_t)cb.m_Results.size(),
       sizeof(VkPerformanceCounterResultKHR) * perfResults.size(), &perfResults[0],
       sizeof(VkPerformanceCounterResultKHR) * counters.size(), VK_QUERY_RESULT_WAIT_BIT);
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   ObjDisp(dev)->DestroyQueryPool(Unwrap(dev), queryPool, NULL);
 
@@ -699,7 +714,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCountersKHR(const rdcarray<GPUCounter
   return ret;
 }
 
-struct VulkanGPUTimerCallback : public VulkanDrawcallCallback
+struct VulkanGPUTimerCallback : public VulkanActionCallback
 {
   VulkanGPUTimerCallback(WrappedVulkan *vk, VulkanReplay *rp, VkQueryPool tsqp, VkQueryPool occqp,
                          VkQueryPool psqp)
@@ -709,9 +724,9 @@ struct VulkanGPUTimerCallback : public VulkanDrawcallCallback
         m_OcclusionQueryPool(occqp),
         m_PipeStatsQueryPool(psqp)
   {
-    m_pDriver->SetDrawcallCB(this);
+    m_pDriver->SetActionCB(this);
   }
-  ~VulkanGPUTimerCallback() { m_pDriver->SetDrawcallCB(NULL); }
+  ~VulkanGPUTimerCallback() { m_pDriver->SetActionCB(NULL); }
   void PreDraw(uint32_t eid, VkCommandBuffer cmd) override
   {
     if(m_OcclusionQueryPool != VK_NULL_HANDLE)
@@ -740,19 +755,37 @@ struct VulkanGPUTimerCallback : public VulkanDrawcallCallback
   void PreDispatch(uint32_t eid, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
   bool PostDispatch(uint32_t eid, VkCommandBuffer cmd) override { return PostDraw(eid, cmd); }
   void PostRedispatch(uint32_t eid, VkCommandBuffer cmd) override { PostRedraw(eid, cmd); }
-  void PreMisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override { PreDraw(eid, cmd); }
-  bool PostMisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override
+  void PreMisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
   {
+    if(flags & ActionFlags::PassBoundary)
+      return;
+    PreDraw(eid, cmd);
+  }
+  bool PostMisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
+  {
+    if(flags & ActionFlags::PassBoundary)
+      return false;
     return PostDraw(eid, cmd);
   }
-  void PostRemisc(uint32_t eid, DrawFlags flags, VkCommandBuffer cmd) override
+  void PostRemisc(uint32_t eid, ActionFlags flags, VkCommandBuffer cmd) override
   {
+    if(flags & ActionFlags::PassBoundary)
+      return;
     PostRedraw(eid, cmd);
   }
   void AliasEvent(uint32_t primary, uint32_t alias) override
   {
     m_AliasEvents.push_back(make_rdcpair(primary, alias));
   }
+  bool SplitSecondary() override { return false; }
+  void PreCmdExecute(uint32_t baseEid, uint32_t secondaryFirst, uint32_t secondaryLast,
+                     VkCommandBuffer cmd) override
+  {
+  }
+  void PostCmdExecute(uint32_t baseEid, uint32_t secondaryFirst, uint32_t secondaryLast,
+                      VkCommandBuffer cmd) override
+  {
+  }
 
   void PreEndCommandBuffer(VkCommandBuffer cmd) override {}
   WrappedVulkan *m_pDriver;
@@ -799,7 +832,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
     ret.append(FetchCountersKHR(vkKHRCounters));
   }
 
-  VkPhysicalDeviceFeatures availableFeatures = m_pDriver->GetDeviceFeatures();
+  VkPhysicalDeviceFeatures availableFeatures = m_pDriver->GetDeviceEnabledFeatures();
 
   VkDevice dev = m_pDriver->GetDev();
 
@@ -829,7 +862,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
   VkQueryPool timeStampPool;
   VkResult vkr =
       ObjDisp(dev)->CreateQueryPool(Unwrap(dev), &timeStampPoolCreateInfo, NULL, &timeStampPool);
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   bool occlNeeded = false;
   bool statsNeeded = false;
@@ -858,23 +891,26 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
   if(availableFeatures.occlusionQueryPrecise && occlNeeded)
   {
     vkr = ObjDisp(dev)->CreateQueryPool(Unwrap(dev), &occlusionPoolCreateInfo, NULL, &occlusionPool);
-    RDCASSERTEQUAL(vkr, VK_SUCCESS);
+    CheckVkResult(vkr);
   }
 
   VkQueryPool pipeStatsPool = VK_NULL_HANDLE;
   if(availableFeatures.pipelineStatisticsQuery && statsNeeded)
   {
     vkr = ObjDisp(dev)->CreateQueryPool(Unwrap(dev), &pipeStatsPoolCreateInfo, NULL, &pipeStatsPool);
-    RDCASSERTEQUAL(vkr, VK_SUCCESS);
+    CheckVkResult(vkr);
   }
 
   VkCommandBuffer cmd = m_pDriver->GetNextCmd();
 
+  if(cmd == VK_NULL_HANDLE)
+    return {};
+
   VkCommandBufferBeginInfo beginInfo = {VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO, NULL,
                                         VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT};
 
   vkr = ObjDisp(dev)->BeginCommandBuffer(Unwrap(cmd), &beginInfo);
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   ObjDisp(dev)->CmdResetQueryPool(Unwrap(cmd), timeStampPool, 0, maxEID * 2);
   if(occlusionPool != VK_NULL_HANDLE)
@@ -883,7 +919,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
     ObjDisp(dev)->CmdResetQueryPool(Unwrap(cmd), pipeStatsPool, 0, maxEID);
 
   vkr = ObjDisp(dev)->EndCommandBuffer(Unwrap(cmd));
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
 #if ENABLED(SINGLE_FLUSH_VALIDATE)
   m_pDriver->SubmitCmds();
@@ -901,7 +937,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
       Unwrap(dev), timeStampPool, 0, (uint32_t)m_TimeStampData.size(),
       sizeof(uint64_t) * m_TimeStampData.size(), &m_TimeStampData[0], sizeof(uint64_t),
       VK_QUERY_RESULT_64_BIT | VK_QUERY_RESULT_WAIT_BIT);
-  RDCASSERTEQUAL(vkr, VK_SUCCESS);
+  CheckVkResult(vkr);
 
   ObjDisp(dev)->DestroyQueryPool(Unwrap(dev), timeStampPool, NULL);
 
@@ -913,7 +949,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
         Unwrap(dev), occlusionPool, 0, (uint32_t)m_OcclusionData.size(),
         sizeof(uint64_t) * m_OcclusionData.size(), &m_OcclusionData[0], sizeof(uint64_t),
         VK_QUERY_RESULT_64_BIT | VK_QUERY_RESULT_WAIT_BIT);
-    RDCASSERTEQUAL(vkr, VK_SUCCESS);
+    CheckVkResult(vkr);
 
     ObjDisp(dev)->DestroyQueryPool(Unwrap(dev), occlusionPool, NULL);
   }
@@ -926,7 +962,7 @@ rdcarray<CounterResult> VulkanReplay::FetchCounters(const rdcarray<GPUCounter> &
         Unwrap(dev), pipeStatsPool, 0, (uint32_t)cb.m_Results.size(),
         sizeof(uint64_t) * m_PipeStatsData.size(), &m_PipeStatsData[0], sizeof(uint64_t) * 11,
         VK_QUERY_RESULT_64_BIT | VK_QUERY_RESULT_WAIT_BIT);
-    RDCASSERTEQUAL(vkr, VK_SUCCESS);
+    CheckVkResult(vkr);
 
     ObjDisp(dev)->DestroyQueryPool(Unwrap(dev), pipeStatsPool, NULL);
   }
