@@ -1,7 +1,7 @@
 /******************************************************************************
 * The MIT License (MIT)
 *
-* Copyright (c) 2019-2020 Baldur Karlsson
+* Copyright (c) 2019-2021 Baldur Karlsson
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
@@ -265,7 +265,11 @@ void VulkanGraphicsTest::Prepare(int argc, char **argv)
           vkh::InstanceCreateInfo(app, enabledLayers, enabledInstExts).next(instInfoNext), NULL,
           &inst);
 
-      if(vkr == VK_SUCCESS)
+      if(vkr != VK_SUCCESS)
+      {
+        TEST_ERROR("Error initialising vulkan instance: %d", vkr);
+      }
+      else
       {
         volkLoadInstance((VkInstance)inst);
 
@@ -353,6 +357,10 @@ void VulkanGraphicsTest::Prepare(int argc, char **argv)
   {                                                                       \
     Avail = "Required physical device feature '" #a "' is not supported"; \
     return;                                                               \
+  }                                                                       \
+  if(optFeatures.a && supported.a)                                        \
+  {                                                                       \
+    features.a = VK_TRUE;                                                 \
   }
 
   CHECK_FEATURE(robustBufferAccess);
@@ -646,6 +654,8 @@ bool VulkanGraphicsTest::Init()
   allocInfo.device = device;
   allocInfo.frameInUseCount = 4;
   allocInfo.pVulkanFunctions = &funcs;
+  if(hasExt(VK_KHR_DEDICATED_ALLOCATION_EXTENSION_NAME) && vmaDedicated)
+    allocInfo.flags |= VMA_ALLOCATOR_CREATE_KHR_DEDICATED_ALLOCATION_BIT;
 
   vmaCreateAllocator(&allocInfo, &allocator);
 
@@ -761,7 +771,7 @@ void VulkanGraphicsTest::FinishUsingBackbuffer(VkCommandBuffer cmd, VkAccessFlag
 
 void VulkanGraphicsTest::Submit(int index, int totalSubmits, const std::vector<VkCommandBuffer> &cmds,
                                 const std::vector<VkCommandBuffer> &seccmds, VulkanWindow *window,
-                                VkQueue q)
+                                VkQueue q, bool sync2)
 {
   if(window == NULL)
     window = mainWindow;
@@ -769,7 +779,7 @@ void VulkanGraphicsTest::Submit(int index, int totalSubmits, const std::vector<V
   if(q == VK_NULL_HANDLE)
     q = queue;
 
-  window->Submit(index, totalSubmits, cmds, seccmds, q);
+  window->Submit(index, totalSubmits, cmds, seccmds, q, sync2);
 }
 
 void VulkanGraphicsTest::Present(VulkanWindow *window, VkQueue q)
@@ -1289,37 +1299,75 @@ void VulkanWindow::Acquire()
 }
 
 void VulkanWindow::Submit(int index, int totalSubmits, const std::vector<VkCommandBuffer> &cmds,
-                          const std::vector<VkCommandBuffer> &seccmds, VkQueue q)
+                          const std::vector<VkCommandBuffer> &seccmds, VkQueue q, bool sync2)
 {
-  VkPipelineStageFlags waitStage = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT;
-
-  VkSubmitInfo submit = vkh::SubmitInfo(cmds);
-
-  if(index == 0)
-  {
-    submit.waitSemaphoreCount = 1;
-    submit.pWaitDstStageMask = &waitStage;
-    submit.pWaitSemaphores = &renderStartSemaphore;
-  }
-
-  if(index == totalSubmits - 1)
-  {
-    submit.signalSemaphoreCount = 1;
-    submit.pSignalSemaphores = &renderEndSemaphore;
-  }
-
   VkFence fence;
   CHECK_VKR(vkCreateFence(m_Test->device, vkh::FenceCreateInfo(), NULL, &fence));
 
   fences.insert(fence);
 
+  if(sync2)
+  {
+    VkSubmitInfo2KHR submit = {VK_STRUCTURE_TYPE_SUBMIT_INFO_2_KHR};
+
+    std::vector<VkCommandBufferSubmitInfoKHR> cmdSubmits;
+    for(VkCommandBuffer cmd : cmds)
+      cmdSubmits.push_back({VK_STRUCTURE_TYPE_COMMAND_BUFFER_SUBMIT_INFO_KHR, NULL, cmd, 0});
+
+    submit.commandBufferInfoCount = (uint32_t)cmdSubmits.size();
+    submit.pCommandBufferInfos = cmdSubmits.data();
+
+    VkSemaphoreSubmitInfoKHR renderStart = {}, renderEnd = {};
+
+    if(index == 0)
+    {
+      renderStart.sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO_KHR;
+      renderStart.semaphore = renderStartSemaphore;
+      renderStart.stageMask = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT_KHR;
+
+      submit.waitSemaphoreInfoCount = 1;
+      submit.pWaitSemaphoreInfos = &renderStart;
+    }
+
+    if(index == totalSubmits - 1)
+    {
+      renderEnd.sType = VK_STRUCTURE_TYPE_SEMAPHORE_SUBMIT_INFO_KHR;
+      renderEnd.semaphore = renderEndSemaphore;
+      renderEnd.stageMask = VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT_KHR;
+
+      submit.signalSemaphoreInfoCount = 1;
+      submit.pSignalSemaphoreInfos = &renderEnd;
+    }
+
+    CHECK_VKR(vkQueueSubmit2KHR(q, 1, &submit, fence));
+  }
+  else
+  {
+    VkPipelineStageFlags waitStage = VK_PIPELINE_STAGE_ALL_COMMANDS_BIT;
+
+    VkSubmitInfo submit = vkh::SubmitInfo(cmds);
+
+    if(index == 0)
+    {
+      submit.waitSemaphoreCount = 1;
+      submit.pWaitDstStageMask = &waitStage;
+      submit.pWaitSemaphores = &renderStartSemaphore;
+    }
+
+    if(index == totalSubmits - 1)
+    {
+      submit.signalSemaphoreCount = 1;
+      submit.pSignalSemaphores = &renderEndSemaphore;
+    }
+
+    CHECK_VKR(vkQueueSubmit(q, 1, &submit, fence));
+  }
+
   for(const VkCommandBuffer &cmd : cmds)
     pendingCommandBuffers[0].push_back(std::make_pair(cmd, fence));
 
   for(const VkCommandBuffer &cmd : seccmds)
     pendingCommandBuffers[1].push_back(std::make_pair(cmd, fence));
-
-  CHECK_VKR(vkQueueSubmit(q, 1, &submit, fence));
 }
 
 void VulkanWindow::Present(VkQueue queue)
@@ -1411,6 +1459,12 @@ void VulkanGraphicsTest::getPhysProperties2(void *nextStruct)
   }
 }
 
+bool VulkanGraphicsTest::hasExt(const char *ext)
+{
+  return std::find_if(devExts.begin(), devExts.end(),
+                      [ext](const char *a) { return !strcmp(a, ext); }) != devExts.end();
+}
+
 template <>
 VkFormat vkh::_FormatFromObj<Vec4f>()
 {
